#include "modem.h"

uint16_t k = 1;

// Server Addresses and Sockets
static int16_t server_send_fd; 
static struct sockaddr_in6 addr_server_sendto;
uint32_t msg_rand_id;

static char client_id_imei[16] = {0};											
static char client_rsrp[5] = {0};				
static char client_rsrp_val = 0;

static uint8_t txbuf[TX_BUFFER_LEN] = {0};
static uint16_t txbuf_len = TX_BUFFER_LEN;

K_SEM_DEFINE(lte_connected, 0, 1);

void rsrp_cb(char rsrp_value)
{
	client_rsrp_val = rsrp_value;
	printk("The RSRP of the system: %d\n", client_rsrp_val);
}

void modem_transmitData(void) {
	memset(txbuf, 0, txbuf_len);
	memcpy(&txbuf[0], client_id_imei, 15);	
	// Initialize all elements of the array
	uint8_t j = 15;
    for (uint16_t i = 15; i < (txbuf_len-1); i++) {
		if (j>255) 
		{
			j = 0;
			printk("\n\n");
		}
    	txbuf[i] = j;
		j = j+1;
	}
	txbuf[txbuf_len-1] = k;
	k = k+1;
    modem_server_transmission_work_fn(NULL);
}

void modem_server_transmission_work_fn(struct k_work *work)
{
	int8_t err;
	printk("\nIP address %s, Port number %d\n",
		       CONFIG_TCP_SERVER_ADDRESS_STATIC,
		       CONFIG_TCP_SERVER_PORT);	

	printk("\n*********************************************\n");
	printk("Data to send over LTE-M (as integers, with IMEI embedded as ASCII):\n");
	for(int i = 0; i < txbuf_len; i++) {
		printk("%d\t", txbuf[i]);
	}
	printk("\n*********************************************\n");
	
	err = send(server_send_fd, txbuf, (size_t)txbuf_len, 0);
	if (err < 0) {
		printk("Failed to transmit TCP packet, %d %s\n", errno, strerror(errno));
	} else {
		printk("Successfully transmitted TCP packet\n");
	}
}

void modem_main_init(void)
{
	int8_t err;
    char imei_buf[20 + sizeof("OK\r\n")];
	char rsrp_buf[10 + sizeof("OK\r\n")];

    printk("Connecting\n");
	
	#if defined(CONFIG_NRF_MODEM_LIB)
		
		modem_modem_init();

		modem_modem_connect();   

		k_sem_take(&lte_connected, K_FOREVER);   
	#endif

	err = modem_TCP_init();
	if (err) {
		printk("Not able to initialize TCP server connection\n");
		return;
	}

	printk("Initialising modem info library...\n");
	err = modem_info_init();
	if (err) {
		printk("Unable to initialise modem info library, error: %d\n", err);
	}

    printk("Querying IMEI\n");
	
	modem_info_string_get(MODEM_INFO_IMEI, imei_buf, sizeof(imei_buf));
	strncpy(client_id_imei, imei_buf, sizeof(client_id_imei) - 1);
	printk("IMEI : %s\n", client_id_imei);
	modem_info_string_get(MODEM_INFO_RSRP, rsrp_buf, sizeof(rsrp_buf));
	strncpy(client_rsrp, rsrp_buf, sizeof(client_rsrp) - 1);			
	printk("RSRP : %s\n", client_rsrp);
}

// **************************************** Function Definition **************************************
#if defined(CONFIG_NRF_MODEM_LIB) 
	
	// To initialize the modem 
	void modem_modem_init(void)
	{
		int8_t err;

		printk("Modem initialization...\n");

		err = nrf_modem_lib_init();
		if (err) {
			printk("Failed to initialize modem library, error: %d\n", err);
			return;
		}

		#if defined(CONFIG_TCP_RAI_ENABLE)
			if (is_nrf9160()) {
				err = nrf_modem_at_printf("AT%%REL14FEAT=0,1,0,0,0");
				if (err) {
					printk("Failed to enable Access Stratum RAI support, error: %d\n", err);
					return err;
				}
			}
		#endif

		err = lte_lc_init();
		if (err) {
			printk("Modem initialization failed, error: %d\n", err);
			return;
		}

		return;
	}


	// Connect MODEM to the Cellular Network
	void modem_modem_connect(void)
	{
		int8_t err;
		if (IS_ENABLED(CONFIG_LTE_AUTO_INIT_AND_CONNECT)) {
		} else {
			err = lte_lc_connect_async(modem_lte_handler);
			if (err) {
				printk("Connecting to LTE network failed, error: %d\n", err);
				return;
			}
		}
	}

	
	void modem_lte_handler(const struct lte_lc_evt *const evt)
	{
		switch (evt->type) {
		case LTE_LC_EVT_NW_REG_STATUS:
			if ((evt->nw_reg_status != LTE_LC_NW_REG_REGISTERED_HOME) &&
	     		(evt->nw_reg_status != LTE_LC_NW_REG_REGISTERED_ROAMING)) {
				break;
			}

			printk("Network registration status: %s\n",
				evt->nw_reg_status == LTE_LC_NW_REG_REGISTERED_HOME ?
				"Connected - home network" : "Connected - roaming\n");
			k_sem_give(&lte_connected);
			break;
		case LTE_LC_EVT_PSM_UPDATE:
			printk("PSM parameter update: TAU: %d, Active time: %d\n",
			evt->psm_cfg.tau, evt->psm_cfg.active_time);
			break;
		case LTE_LC_EVT_EDRX_UPDATE: {
			char log_buf[60];
			ssize_t len;
			len = snprintf(log_buf, sizeof(log_buf),
		       		"eDRX parameter update: eDRX: %f, PTW: %f\n",
		       		evt->edrx_cfg.edrx, evt->edrx_cfg.ptw);
			if (len > 0) {
				printk("%s\n", log_buf);
			}
			break;
		}
		case LTE_LC_EVT_RRC_UPDATE:
			printk("RRC mode: %s\n",
				evt->rrc_mode == LTE_LC_RRC_MODE_CONNECTED ?
				"Connected" : "Idle\n");
			break;
		case LTE_LC_EVT_CELL_UPDATE:
			printk("LTE cell changed: Cell ID: %d, Tracking area: %d\n",
	       		evt->cell.id, evt->cell.tac);
			break;
		default:
			break;
		}
	}
#endif

int8_t modem_TCP_init(void)
{
	int8_t err;
	int8_t status;

	// Initialise random number based message ID
	msg_rand_id = sys_rand32_get();

	addr_server_sendto.sin6_family = AF_INET6;
	addr_server_sendto.sin6_port = htons(CONFIG_TCP_SERVER_PORT);
	status = inet_pton(AF_INET6, CONFIG_TCP_SERVER_ADDRESS_STATIC, &addr_server_sendto.sin6_addr);
	if (status == 0) {
		printk("src does not contain a character string representing a valid network address\n");
		return -1;
	} else if(status < 0) {
		printk("inet_pton failed: %d %s\n", errno, strerror(errno));
		err = -errno;
		goto error;
	}
	
	// Set up server socket and connect
	server_send_fd = socket(AF_INET6, SOCK_STREAM, IPPROTO_TCP);
	// server_send_fd = socket(AF_INET6, SOCK_STREAM, 0);
	if (server_send_fd < 0) {
		printk("TCP socket creation failed: %d\n", errno);
		err = -errno;
		goto error;
	}

	err = connect(server_send_fd, (struct sockaddr *)&addr_server_sendto, sizeof(struct sockaddr_in6));
	if (err < 0) {
		printk("Connect failed : %d\n", errno);
		goto error;
	}

	return 0;
error:
	modem_server_disconnect();
	return err;
}

void modem_server_disconnect(void)
{
	(void)close(server_send_fd);
	printk("Closed the socket\n");
}

